#' @name block
#' @rdname block
#' @title Slice sampling in blocks
#'
#' @description Handle the case where different blocks of the random
#' vector have different scales.  This is a wrapper for
#' \code{\link{slice}}.
#'
#' @param blocks list of blocks, forming a partition of \code{1:d}
#' @param scale initial scale, either common or one for each block
#' @param m maximum number of stepping out steps, can be \code{Inf} (same for each block)
#' @param rand process blocks in random order?  See Details.
#'
#' @return \code{blockslice} and \code{ooatslice} have the same return
#' value as \code{\link{slice}}.
#'
#' \code{getscales} is a utility function to extract the scales from a
#' function returned by \code{slice} (scalar), \code{blockslice} (one
#' per block), or \code{ooatslice} (one per component).
#'
#' @seealso \code{\link{slice}} for explanations of the \code{scale}
#' and \code{m} arguments.
#'
#' @author Jonathan Rougier <j.c.rougier@@bristol.ac.uk>
#' 
#' @references J. Besag, P. Green, D. Higdon, and K. Mengersen, 1995,
#' Bayesian Computation and Stochastic Systems, Statistical Science
#' 10(1), 3-66.
#' 
#' 
#' @examples
#' ## Similar example to 'slice', but making a weird distribution with
#' ## identical margins
#' 
#' mylogf <- function(x) {
#'   sum(dnorm(x, log = TRUE)) + log(sum(x) <= 0)
#' }
#' 
#' myslice <- blockslice(list(1, c(2, 4), c(3, 5, 6))) # or oaatslice(6) for individual scales
#' N <- 1e3
#' rsam <- matrix(NA, N, 6)
#' spinup <- 1e2
#' 
#' x <- rep(-1, 6)
#' for (loop in (-spinup + 1):N) {
#'   x <- myslice(x, mylogf, learn = (loop <= 0))
#'   if (loop > 0) rsam[loop, ] <- x
#' }
#' print(all(rowSums(rsam) <= 0)) # TRUE
#' pairs(rsam, xlim = c(-3, 3), ylim = c(-3, 3))
#'
#' ## check the margins against each other (nb blocks are /not/ symmetrical)
#'
#' mvqqplot(rsam, rsam[, c(2:6, 1)])
#' mvqqplot(rsam, rsam[, c(2:6, 1)], list(2:3, 2:4, 2:5, 2:6, 1:6))
#' 
#' ## comparison sample by rejection sampling
#' 
#' rsam1 <- matrix(rnorm(2 * N * 6), ncol = 6)
#' keep <- rowSums(rsam1) <= 0
#' rsam1 <- rsam1[keep, ]
#' 
#' mvqqplot(rsam, rsam1)
#' mvqqplot(rsam, rsam1[, c(2:6, 1)])
#' mvqqplot(rsam, rsam1, list(2, 2:3, 2:4, 1:4))
#' 
#' ## looks OK!
#'
#' print(getscales(myslice)) # should be about the same
#' 
#' #### sample from a level set of the constrained banana function
#' 
#' mylogf <- function(x, theta) {
#'     log(banana(x, theta) <= 100) +
#'         log(-1.5 <= x[1] && x[1] <= 2.0) +
#'             log(-1.0 <= x[2] && x[2] <= 3)
#' }
#' 
#' x <- seq(from = -2, to = 2.5, length = 101)
#' y <- seq(from = -1.5, to = 3.5, length = 101)
#' z <- apply(as.matrix(expand.grid(x, y)), 1, mylogf, theta = 100)
#' dim(z) <- c(length(x), length(y))
#' image(x, y, z)
#' 
#' myslice <- oaatslice(2)
#' N <- 1e3
#' rsam <- matrix(NA, N, 2)
#' spinup <- 1e2
#' 
#' x <- rep(0, 2)
#' for (loop in (-spinup + 1):N) {
#'     x <- myslice(x, mylogf, learn = (loop <= 0), theta = 100)
#'     if (loop > 0) rsam[loop, ] <- x
#' }
#' points(rsam, pch = 20, cex = 0.7)
NULL

#' @rdname block
#' @details If \code{rand = TRUE}, the blocks are processed in random
#' order, which ensures that the sampler is reversible. Reversible
#' samplers have more attractive theoretical (convergence) properties;
#' see Besag et al (1995).
#' 
#' @export

blockslice <- function(blocks, scale = 1, m = Inf, rand = TRUE) {

  stopifnot(is.partition(blocks), is.logical(rand))
  k <- length(blocks)
  d <- max(unlist(blocks))
  stopifnot(length(scale) %in% c(1, k))
  if (length(scale) == 1) scale <- rep(scale, k)

  if (d == 1 && rand)
      warning("\'rand = TRUE\' has no effect when d = 1")

  ## set up a slice function for each block
  
  slicelist <- lapply(1L:k, function(i)
                      slice(d = length(blocks[[i]]), scale = scale[i], m = m))

  ## now create a block slicing function as a Gibbs sampler

  function(x0, logf, learn = FALSE, ...) {

      logf1 <- function(x) logf(x, ...)
      iset <- 1L:k
      if (rand && k > 1) iset <- sample(iset)
      
    x1 <- x0
    for (i in iset) {
      bb <- blocks[[i]]
      foo <- with(list(x1 = x1, bb = bb, logf1 = logf1),
                  function(x) {
                      x1[bb] <- x
                      logf1(x1)
                  })
      x1[bb] <- slicelist[[i]](x0[bb], foo, learn = learn)
    }

    x1
  }
}

## here is another useful function (local)

is.partition <- function(x) {
  if (!is.list(x)) return(FALSE)
  x <- unlist(x)
  d <- max(x)
  !anyDuplicated(x) && setequal(x, 1:d)
}

#' @rdname block
#' @param d number of components in the random vector
#' @param ... arguments for \code{blockslice}
#' @details \code{oaatslice} is just a wrapper for \code{blockslice}, where each component is in a block of one ('one at a time').  This allows each component to have its own scale.  But this approach will not respond very well in situations where two or more components with similar scales are strongly correlated (in which case they should be in the same block).
#'
#' @export

## just a wrapper for blockslice

oaatslice <- function(d, ...) {
  stopifnot(length(d) == 1, is.wholenumber(d), d > 0)
  blockslice(as.list(1:d), ...)
}

#' @rdname block
#' @param obj Function returned by \code{\link{slice}}, \code{blockslice}, or \code{oaatslice}
#' @export
#'

getscales <- function(obj) {
    stopifnot(is.function(obj))
    env <- environment(obj)
    if (exists("blocks", envir = env)) {
        slicelist <- get("slicelist", envir = env)
        sapply(slicelist, getscales)
    } else get("scale", envir = env)
}
