#' Slice sampler
#' 
#' Adaptive multivariate slice sampler.  The univariate sampler
#' implements the Neal (2003) stepping out method.  The multivariate
#' version uses univariate sampling in a random direction.  Adaptation
#' is via a learning phase to tune the stepping-out scale.  Only a single
#' scale is used: for blocks of components with different scales, use
#' \code{\link{blockslice}}.
#'
#' @param d dimension of the random vector
#' @param scale  An initial scale which is common across all components (see Details)
#' @param m maximum number of stepping out steps, can be \code{Inf} (see Details)
#'
#' @return A slice-sampling function to go into an MCMC iteration.
#' This has arguments
#' \tabular{rl}{
#' x0: \tab the current point \cr\cr
#' logf: \tab logarithm of the target distribution (up to an additive constant) \cr\cr
#' learn: \tab logical to control learning \cr\cr
#' ...: \tab additional arguments to \code{logf}.
#' }
#' The return value is a new point.
#' 
#' @details
#'
#' The \code{scale} parameter controls the efficiency of the
#' sampler. It should be something similar to the interquartile range
#' averaged over the set of full conditional distributions.  Make a
#' reasonable guess, and then set \code{learn = TRUE} for a while in
#' your sampler (eg during spin-up, see Examples), so that this value
#' is adapted effectively.  The scale can be extracted using
#' \code{\link{getscales}}.
#' 
#' Neal's stepping out algorithm provides for an upper bound on the
#' number of steps, which may be infinite.  This is controlled by the
#' argument \code{m}.  While learning, \code{m = Inf} to ensure that
#' the step sizes best reflect the nature of the target distribution.
#' But once learning has been done, \code{m} reverts to its supplied
#' value (which may be \code{Inf}) which may speed up the algorithm,
#' but at the expense of slower mixing.  If you are getting impatient,
#' you might try a value such as \code{m = 20} to see if that makes a
#' difference.
#'
#' @seealso \code{\link{blockslice}} for when the length scales need
#' to differ, \code{\link{getscales}} for extracting the scales.  \code{\link{mvqqplot}}
#' for checking the result against a different simulation method (see Example below).
#' 
#' @references R.M. Neal, 2003, Slice Sampling, The Annals of
#' Statistics, 31(3), 705-741.
#'
#' @author Jonathan Rougier <j.c.rougier@@bristol.ac.uk>
#' 
#' @export
#' @examples
#' 
#' ## sample from the distribution X[1:4] IID N(0, 1) st sum(X) <= 0 & X[1:2] <= 0
#' 
#' mylogf <- function(x) {
#'   sum(dnorm(x, log = TRUE)) + log(sum(x) <= 0 & all(x[1:2] <= 0))
#' }
#' 
#' myslice <- slice(4)
#' N <- 1e3
#' rsam <- matrix(NA, N, 4)
#' spinup <- 1e2
#' 
#' x <- rep(-1, 4)
#' for (loop in (-spinup + 1):N) {
#'   x <- myslice(x, mylogf, learn = (loop <= 0))
#'   if (loop > 0) rsam[loop, ] <- x
#' }
#' print(all(rowSums(rsam) <= 0)) # TRUE
#' pairs(rsam, xlim = c(-3, 3), ylim = c(-3, 3))
#' 
#' ## comparison sample by rejection sampling
#' 
#' rsam1 <- matrix(rnorm(4 * N * 4), ncol = 4)
#' keep <- rowSums(rsam1) <= 0 & apply(rsam1[, 1:2] <= 0, 1, all)
#' rsam1 <- rsam1[keep, ]
#' pairs(rsam1, xlim = c(-3, 3), ylim = c(-3, 3))
#' 
#' mvqqplot(rsam, rsam1)
#' mvqqplot(rsam, rsam1, list(2, 2:3, 2:4, 1:4))
#' 
#' ## looks OK!

slice <- function(d, scale = 1, m = Inf) {

    stopifnot(length(d) == 1L, is.wholenumber(d), d > 0)
    d <- as.integer(d)

    stopifnot(length(scale) == 1L, is.finite(scale), scale > 0,
              length(m) == 1L, m == Inf || is.wholenumber(m), m > 0)

    W <- rep(NA, 1000L)
    n <- 0L
    updated <- FALSE

    function(x0, logf, learn = FALSE, ...) {

        logf1 <- function(x) logf(x, ...)
        stopifnot(length(x0) == d)
      
        ## this is where updating happens, once only

        if (!updated && !learn) {
            if (n == 0) warning("Using default scale", immediate. = TRUE)
            else scale <<- mean(abs(W), na.rm = TRUE)
            updated <<- TRUE
        }

        w <- scale
        if (!updated) mm <- Inf else mm <- m # always have m = Inf when learning

        ## get a random direction

        if (d > 1L) {
            angle <- rnorm(d)
            angle <- angle / sum(angle^2)
            imax <- which.max(abs(angle))[1L]
            if (angle[imax] < 0)
                angle <- -1 * angle # so +ve w goes to the right in x0[imax]
            w <- w * angle # now a vector
        } else imax <- 1L

        ##  ## ## ## ## ## This code based on Neal (2003) sec 4

        ## initial random interval
  
        y <- log(runif(1)) + logf1(x0)
        if (!is.finite(y)) stop("inappropriate \'x0\' value")
        U <- runif(1)
        L <- x0 - w * U
        R <- L + w

        ## expand interval: using mm, local value of m

        if (mm < Inf) {
          V <- runif(1)
          J <- floor(mm * V)
          K <- (mm - 1) - J
          if (J > 0) for (j in J:1) {
            L <- L - w
            if (y >= logf1(L)) break
          }
          if (K > 0) for (k in K:1) {
            R <- R + w
            if (y >= logf1(R)) break
          }
        } else { # mm == Inf
          while (y < logf1(L)) L <- L - w
          while (y < logf1(R)) R <- R + w
        }

        ## sample uniformly and shrink interval to get x1

        repeat {
            U <- runif(1)
            x1 <- L + U * (R - L)
            if (y < logf1(x1)) break
            if (x1[imax] < x0[imax]) L <- x1 else R <- x1
        }

        ## ## ## ## ## ##
        
        ## can learn about the stepsize

        if (!updated) {
            n <<- n + 1L
            if (n > length(W))
                W <<- c(W, rep(NA, 1000L))
            W[n] <<- as.vector(dist(rbind(x0, x1)))
        }

        ## return the new point
      
        x1
    }
}

## this useful function used in blockslice as well (local)

is.wholenumber <- function(x)
  is.finite(x) & x == floor(x)
