% Generated by roxygen2 (4.0.1): do not edit by hand
\name{slice}
\alias{slice}
\title{Slice sampler}
\usage{
slice(d, scale = 1, m = Inf)
}
\arguments{
\item{d}{dimension of the random vector}

\item{scale}{An initial scale which is common across all components (see Details)}

\item{m}{maximum number of stepping out steps, can be \code{Inf} (see Details)}
}
\value{
A slice-sampling function to go into an MCMC iteration.
This has arguments
\tabular{rl}{
x0: \tab the current point \cr\cr
logf: \tab logarithm of the target distribution (up to an additive constant) \cr\cr
learn: \tab logical to control learning \cr\cr
...: \tab additional arguments to \code{logf}.
}
The return value is a new point.
}
\description{
Adaptive multivariate slice sampler.  The univariate sampler
implements the Neal (2003) stepping out method.  The multivariate
version uses univariate sampling in a random direction.  Adaptation
is via a learning phase to tune the stepping-out scale.  Only a single
scale is used: for blocks of components with different scales, use
\code{\link{blockslice}}.
}
\details{
The \code{scale} parameter controls the efficiency of the
sampler. It should be something similar to the interquartile range
averaged over the set of full conditional distributions.  Make a
reasonable guess, and then set \code{learn = TRUE} for a while in
your sampler (eg during spin-up, see Examples), so that this value
is adapted effectively.  The scale can be extracted using
\code{\link{getscales}}.

Neal's stepping out algorithm provides for an upper bound on the
number of steps, which may be infinite.  This is controlled by the
argument \code{m}.  While learning, \code{m = Inf} to ensure that
the step sizes best reflect the nature of the target distribution.
But once learning has been done, \code{m} reverts to its supplied
value (which may be \code{Inf}) which may speed up the algorithm,
but at the expense of slower mixing.  If you are getting impatient,
you might try a value such as \code{m = 20} to see if that makes a
difference.
}
\examples{
## sample from the distribution X[1:4] IID N(0, 1) st sum(X) <= 0 & X[1:2] <= 0

mylogf <- function(x) {
  sum(dnorm(x, log = TRUE)) + log(sum(x) <= 0 & all(x[1:2] <= 0))
}

myslice <- slice(4)
N <- 1e3
rsam <- matrix(NA, N, 4)
spinup <- 1e2

x <- rep(-1, 4)
for (loop in (-spinup + 1):N) {
  x <- myslice(x, mylogf, learn = (loop <= 0))
  if (loop > 0) rsam[loop, ] <- x
}
print(all(rowSums(rsam) <= 0)) # TRUE
pairs(rsam, xlim = c(-3, 3), ylim = c(-3, 3))

## comparison sample by rejection sampling

rsam1 <- matrix(rnorm(4 * N * 4), ncol = 4)
keep <- rowSums(rsam1) <= 0 & apply(rsam1[, 1:2] <= 0, 1, all)
rsam1 <- rsam1[keep, ]
pairs(rsam1, xlim = c(-3, 3), ylim = c(-3, 3))

mvqqplot(rsam, rsam1)
mvqqplot(rsam, rsam1, list(2, 2:3, 2:4, 1:4))

## looks OK!
}
\author{
Jonathan Rougier <j.c.rougier@bristol.ac.uk>
}
\references{
R.M. Neal, 2003, Slice Sampling, The Annals of
Statistics, 31(3), 705-741.
}
\seealso{
\code{\link{blockslice}} for when the length scales need
to differ, \code{\link{getscales}} for extracting the scales.  \code{\link{mvqqplot}}
for checking the result against a different simulation method (see Example below).
}

